/*********************************************************************/
/*                  pC++/Sage++  Copyright (C) 1993                  */
/*  Indiana University  University of Oregon  University of Rennes   */
/*********************************************************************/


# include <stdio.h>
# include <stdlib.h>
# include "sage++user.h"
# include "sage++callgraph.h"
# include "regexp.h"
# include "instr.h"

// including files for breakpoint debugging.;
#include "baseClasses.h"
#include "sageUtils.h"
#include "sage++classhierarchy.h"
#include "TItypeDesc.h"
  
#ifndef __GNUC__
#  include <unistd.h>
#else
extern "C" int unlink(const char *path);
extern "C" int access(char *path, int amode);
#endif

#include "compatible.h"
#ifdef SYS5
#include <string.h>
#else
#include <strings.h>
#endif

# ifndef R_OK
#   define R_OK 04
# endif

/* -- external functions -------------------------------------------------- */
/* -- from: readcomfile.y ------------------------------------------------- */
extern "C" int readComFile (char *filename);

/* -- from: namecheck.c --------------------------------------------------- */
extern "C" void initCommands (void);
extern "C" int isValidFunc (char *filename, char *classname, char *funcname);
extern "C" int isNotTooSmall (int numstmts, int numchild);

/* -- from: gendescr.c ---------------------------------------------------- */
extern "C" int newEventId (char *classname, char *funcname, int tag);
extern "C" void genTdl (char *tdlfile);
extern "C" void genEdf (char *edffile);
extern "C" void genFuncTable (char *tabfile, char *depfile, int* pcxx_tag);

/* -- internal functions -------------------------------------------------- */
static void instrFunctions (SgFile *, char *inFile, char *outFile);
static void instrument_coll_const (SgSymbol * fsymb, SgFile *file);

/* -- global variables ---------------------------------------------------- */
int verboseMode = FALSE;
int profileMode = FALSE;
int debugMode = FALSE;
int pcxx_tag[10];
int breakpointMode = FALSE;
static int numfunc = 0;

void printUsage (char* execName)
{
  fprintf (stderr,"Usage: %s\n",execName);
  fprintf (stderr,"    [-help | -h] [-usage | -u] [-debug]\n");
  fprintf (stderr,"    [-verbose | -v] [(-commands | -c) comfile]\n");
  fprintf (stderr,"    [(-tdl | -t) tdlfile] [(-o | -output) outfile]\n");
  fprintf (stderr,"    [(-edf | -e) edffile] [-profile | -p]\n");
  fprintf (stderr,"    [(-breakpoint | -b) ]\n");
  fprintf (stderr,"    depfile\n");
  fprintf (stderr,
     "\n'%s -help' prints description of switches, parameters\n", execName);
  exit (1);
}

void printHelp (char* execName)
{
  printf ("Usage: %s\n", execName);
  printf ("  [-h | -help]  Print this help text\n");
  printf ("  [-u | -usage]  Print usage information (command line switches)\n");
  printf ("  [-v | -verbose]  Turn on Verbose Mode \n");
  printf ("  [-debug]  Print useful debugging information\n");
  printf ("  [(-c | -commands) comfile]  use instrumentation command file\n");
  printf ("  [(-t | -tdl) tdlfile]  Generate TDL description file\n");
  printf ("  [(-e | -edf) edffile]  Generate event descriptor file\n");
  printf ("  [(-o | -output) outfile]  Output dep file name\n");
  printf ("  [-p | -profile]  instrument for profiling\n");
  printf ("  [-b | -breakpoint]  instrument for breakpoint executive\n");
  printf ("  depfile\n");
  exit (1);
}

main (int argc, char **argv)
{
  SgProject *project = 0;
  char *projFile = NULL;
  char *edfFile = NULL;
  char *tdlFile = NULL;
  char *outFile = NULL;
  char *depFile = NULL;
  char *p;
  FILE *pf = 0;
  int l;
  // for breakpoint code;
  classHierarchy *hierarchies ;
 
  if (argc == 1)
  {
    printUsage (argv[0]);
    exit (1);
  }
  else
  {
    initCommands ();
    for(int i = 1; i < argc; i++)
    {
      if ( argv[i][0] == '-' )
      {
        p = argv[i];
        if ( (!strcmp(p,"-help")) || (!strcmp(p,"-h")) )
        {
          printHelp (argv[0]);
        }
        else if ( (!strcmp(p,"-usage")) || (!strcmp(p,"-u")) )
        {
          printUsage (argv[0]);
        }
        else if ( !strcmp(p,"-debug") )
        {
          debugMode = TRUE;
        }
        else if ( (!strcmp(p,"-verbose")) || (!strcmp(p,"-v")) )
        {
          verboseMode = TRUE;
        }
        else if ( (!strcmp(p,"-profile")) || (!strcmp(p,"-p")) )
        {
          profileMode = TRUE;
	  numfunc = 10;
        }
        else if ( (!strcmp(p,"-breakpoint")) || (!strcmp(p,"-b")) )
        {
          breakpointMode = 1;
          if ( i == (argc-1) )
          {
            fprintf (stderr,"ERROR: not enough arguments for -b switch\n");
            printUsage (argv[0]);
          }
          i++;
        }
        else if ( (!strcmp(p,"-commands")) || (!strcmp(p,"-c")) )
        {
          if ( i == (argc-1) )
          {
            fprintf (stderr,"ERROR: not enough arguments for -c switch\n");
            printUsage (argv[0]);
          }
          i++;
          if ( verboseMode ) printf ("reading commands file %s\n", argv[i]);
          if ( ! readComFile (argv[i]) ) exit (1);
        }
        else if ( (!strcmp(p,"-output")) || (!strcmp(p,"-o")) )
        {
          if ( i == (argc-1) )
          {
            fprintf (stderr,"ERROR: not enough arguments for -o switch\n");
            printUsage (argv[0]);
          }
          i++;
          outFile = argv[i];
        }
        else if ( (!strcmp(p,"-tdl")) || (!strcmp(p,"-t")) )
        {
          if ( i == (argc-1) )
          {
            fprintf (stderr,"ERROR: not enough arguments for -t switch\n");
            printUsage (argv[0]);
          }
          i++;
          tdlFile = argv[i];
        }
        else if ( (!strcmp(p,"-edf")) || (!strcmp(p,"-e")) )
        {
          if ( i == (argc-1) )
          {
            fprintf (stderr,"ERROR: not enough arguments for -e switch\n");
            printUsage (argv[0]);
          }
          i++;
          edfFile = argv[i];
        }
        else
        {
          fprintf (stderr,"ERROR: unknown option `%s'\n", p);
          printUsage (argv[0]);
        }
      }
      else
      {
        depFile = argv[i];
        l = strlen(argv[i]);
        p = argv[i] + l - 4;
        if ( strcmp (p, ".dep") != 0 )
        {
          fprintf (stderr,"ERROR: `%s': no dep file\n", argv[i]);
          exit (1);
        }
        else
        {
          *p = '\0';
          projFile = new char[l+2];
          sprintf (projFile, "%s.proj", argv[i]);
          *p = '.';
        }

        if ( access (projFile, R_OK) == -1 )
        {
          pf = fopen (projFile, "w");
          fprintf (pf, "%s\n", depFile);
          fclose (pf);
        }

        if ( verboseMode || debugMode )
          printf ("using project file %s\n", projFile);
        project = new SgProject (projFile);
      }
    }
  }

  if ( depFile )
  {
    if ( ! outFile ) outFile = depFile;

    if ( verboseMode || debugMode )
      printf ("processing %s:\n", project->fileName(0));

    
    // -----------------------------------------------------------
    // creating type description file;
    if (breakpointMode) {

      // initialize sage items for breakpointing..;
      brk_mySageInit();

      // generate type description file..;
      hierarchies = new classHierarchy();

      // declare sage composite lists..;
      brk_sCompositeList *sElemList, *sElemTypeList;
      
      // collectionElemLIST and elementTypeLIST are globals;
      //   they represent the list of all collections and;
      //   the list of all collection element types respectively;
      
      sElemList = (brk_sCompositeList *) brk_collectionElemLIST;
      sElemTypeList = (brk_sCompositeList *) brk_elementTypeLIST;
      
      SgFile *f;
      for (int j=0; j<project->numberOfFiles(); j++) {
	if (debugMode) {
	  printf("determing class hierarchy for %s\n", project->fileName(j));
	}
	f = &(project->file(j));
	hierarchies->determineClassHierarchy(f);
	brk_buildTypesFromFile(f);
      }

      brk_determineSuperClasses(hierarchies);
      if (debugMode) {
	printf("determined super classes\n");
      }
      fflush(stdout);

      for (j=0; j<project->numberOfFiles(); j++) {
	f = &(project->file(j));
	if (debugMode) {
	  printf("instr processing sElemTypeList\n");
	  sElemTypeList->print(0);
	  fflush(stdout);
	}
	sElemTypeList->writeAccessFunctions(f);

	if (debugMode) {
	  printf("instr processing sElemList\n");
	  sElemList->print(0);
	  fflush(stdout);
	}
	sElemList->writeAccessFunctions(f);
	if (debugMode) {
	  printf("done writing access functions for file\n");

	  fflush(stdout);
	}
      
      }

      // generate
      // type description file and access functions.
      // write type description file...;
      brk_printTIFFile();
    }
    
    // -----------------------------------------------------------;

    instrFunctions (&(project->file(0)), project->fileName(0), outFile);

    if ( pf ) unlink(projFile);
  }

  if ( profileMode )
  {
    pcxx_tag[1] = -1;
    pcxx_tag[2] = -1;
    pcxx_tag[3] = -1;

    if ( verboseMode ) printf ("generating functablefile profile.ftab\n");
    SgSymbol *bfunc = (project->file(0)).firstSymbol();
    while ( bfunc )
    {
      if ( isSgFunctionSymb(bfunc) )
      {
        if ( strcmp (bfunc->identifier(), "pcxx_Barrier") == 0 )
          pcxx_tag[1] = bfunc->id();
        else if ( strcmp (bfunc->identifier(), "pcxx_Poll") == 0 )
          pcxx_tag[2] = bfunc->id();
        else if ( strcmp (bfunc->identifier(), "WhereIsData") == 0 )
          pcxx_tag[3] = bfunc->id();
      }
      bfunc = bfunc->next();
    }
    genFuncTable ("profile.ftab", outFile, pcxx_tag);
  }
  else
  {
    if ( tdlFile )
    {
      if ( verboseMode ) printf ("generating TDL file %s\n", tdlFile);
      genTdl (tdlFile);
    }

    if ( edfFile )
    {
      if ( verboseMode ) printf ("generating edf file %s\n", edfFile);
      genEdf (edfFile);
    }
  }
  exit (0);
  return 0; //to make the compiler happy;
}

void instrFunctions (SgFile *file, char *inFile, char *outFile)
{
  SgCallGraph CG;
  SgDerivedType *t;

  if ( file->languageType() == CSrc )
  {
    /* -- Initialize call graph module                             -- */
    /* -------------------------------------------------------------- */
    CG.GenCallTree(file);
    CG.computeClosures();

    if (!breakpointMode) {

      /* -- Search for class Profiler symbol                         -- */
      /* -------------------------------------------------------------- */
      SgSymbol *pclass = file->firstSymbol();
      while ( pclass )
	{
	  if ( isSgClassSymb(pclass) &&
	      (strcmp (pclass->identifier(), "Profiler") == 0) )
	    break;
	  pclass = pclass->next();
	}
      if ( !pclass )
	{
	  fprintf (stderr,"ERROR: `%s': class Profiler declaration missing\n",
               inFile);
	  exit (1);
	}
      t = new SgDerivedType(*pclass);
    }

    /* -- Declarations for instrumenting functions                 -- */
    /* -------------------------------------------------------------- */
    SgStatement *func, *mainfunc, *s;
    SgSymbol *funcsymb, *classsymb, *profsymb, *breakpointsymb = NULL;
    SgMemberFuncSymb *memsymb;
    char *funcname, *classname, *filename;
    SgExpression *e1, *e2, *e3, *e, *args;
    int funcline;
    int isMoE, isColl;
    int numchild, numstmts, res;

    /* -- skip through all functions off the file                  -- */
    /* -------------------------------------------------------------- */
    for (int i=0; i<file->numberOfFunctions(); i++)
    {
      /* -- determine type and "size" of function                  -- */
      /* ------------------------------------------------------------ */
      func = file->functions(i);
      funcsymb = func->symbol();
      funcname = funcsymb->identifier();
      funcline = func->lineNumber();
      filename = func->fileName();
      isMoE    = 0;

      if ( memsymb = isSgMemberFuncSymb(funcsymb) )
      {
        classsymb = memsymb->className();
        classname = classsymb->identifier();
        if ( (isColl = (classsymb->variant() == COLLECTION_NAME)) )
          isMoE  = memsymb->isMethodOfElement();

        if ( debugMode )
          printf ("\t%4d:  %s%s%s::%s", funcline, (isColl ? "[Coll] " : ""),
                 (isMoE ? "[MoE] " : ""), classname, funcname);
      }
      else
      {
        classname = 0;
        isColl    = 0;

        if ( debugMode )
          printf ("\t%4d:  %s", funcline, funcname);
      }

      numchild = CG.numberOfFunsCalledFrom(funcsymb);
      numstmts = 0;
      s = func;
      while ( s != func->lastNodeOfStmt() )
      {
        numstmts++;
        s = s->lexNext();
      }
      if ( debugMode ) printf (" <%ds/%dc>\n", numstmts, numchild);

      /* -- check whether function should be instrumented          -- */
      /* ------------------------------------------------------------ */
      if ( profileMode && (strcmp (funcname, "Processor_Main") == 0) )
      {
        mainfunc    = func;
        pcxx_tag[0] = funcsymb->id();
      }
      else if ( strcmp (funcname, "PROFILE_COLL_NAMES") == 0 )
      {
        profsymb = funcsymb;
      }
      else if (breakpointMode && 
	       strcmp (funcname, "brk_QUE_COLLECTION_INSTANCE") == 0 ) {
	if (debugMode) {
	  printf("found add_collection_instance symbol %s\n",
		 funcname);
	}
	breakpointsymb = funcsymb;
      }
      else if (!breakpointMode &&
	       isValidFunc (filename, classname, funcname) &&
                (res = isNotTooSmall (numstmts, numchild)) )
      {
        if ( debugMode ) printf ("\n");
        if ( verboseMode )
        {
          if ( classname )
            printf ("\tinstrumenting %s::%s", classname, funcname);
          else
            printf ("\tinstrumenting %s", funcname);
          if ( res == COUNT_ONLY )
            printf (" [counting only]\n");
          else
            printf ("\n");
        }

        /* -- declare variable pcxx_P of type class Profiler       -- */
        /* ---------------------------------------------------------- */
        SgSymbol *pcxx_P = new SgVariableSymb("pcxx_P", *t);
        pcxx_P->declareTheSymbol(*func);

        /* -- build 1. argument for constructor (eventID)          -- */
        /* ---------------------------------------------------------- */
        SgVariableSymb IDsymb("Ident");
        numfunc = newEventId (classname, funcname, funcsymb->id());
        e1 = new SgValueExp(numfunc);

        if ( ! profileMode )
        {
          /* -- build 2. Argument for MemberOfElement functions    -- */
          /* -- normal functions get argument "0".                 -- */
          /* -------------------------------------------------------- */
          if ( isMoE )
            e2 = new SgVarRefExp(IDsymb);
          else
            e2 = new SgValueExp(0);

          /* -- build 3. Argument: id of function symbol           -- */
          /* -------------------------------------------------------- */
          e3 = new SgValueExp(funcsymb->id());

          e = new SgExprListExp(*e3);
          args = &SgExprListOp(*e2, *e);
          args = &SgExprListOp(*e1, *args);
        }
        else if ( res == COUNT_ONLY ) {
          /* -- build 2. Argument (constant char) so that          -- */
          /* -- "count only" constructor is used.                  -- */
          /* -------------------------------------------------------- */
          e2 = new SgValueExp('C');
          e = new SgExprListExp(*e2);
          args = &SgExprListOp(*e1, *e);
        }
        else
          args = new SgExprListExp(*e1);

        /* -- finish building "pcxx_P" constructor arguments and   -- */
        /* -- add them                                             -- */
        /* ---------------------------------------------------------- */
        SgExpression *classinit = new SgExpression(CLASSINIT_OP); 
        classinit->setRhs(*args); 
        SgExpression *pcxx_P_ref = new SgVarRefExp(*pcxx_P);
        classinit->setLhs(*pcxx_P_ref);
        SgStatement *st = func->lexNext();
        SgExpression *list = st->expr(0);
        list->setLhs(*classinit);
      }
      else if ( debugMode )
      {
        printf ("\t++++++++++++++++++++ NO\n");
      }
    }

    if ( profileMode )
    {
      /* -- instrument Processor_Main: 1.arg (0), 2.arg (numfunc)  -- */
      /* -- in the same way                                        -- */
      /* ------------------------------------------------------------ */
      SgSymbol *pcxx_P = new SgVariableSymb("pcxx_P", *t);
      pcxx_P->declareTheSymbol(*mainfunc);
      e1 = new SgValueExp(0);                // 0
      e2 = new SgValueExp(numfunc+1);        // n (= numfunc+1)
      e = new SgExprListExp(*e2);            // (n)
      args = &SgExprListOp(*e1, *e);         // (0, n)
      SgExpression *classinit = new SgExpression(CLASSINIT_OP); 
      classinit->setRhs(*args); 
      SgExpression *pcxx_P_ref = new SgVarRefExp(*pcxx_P);
      classinit->setLhs(*pcxx_P_ref);
      SgStatement *st = mainfunc->lexNext();
      SgExpression *list = st->expr(0);
      list->setLhs(*classinit);
    }


    // code for breakpoint debugger instrumentation...
    if ( breakpointMode)
    {
      // only when breakpointMode == 2 do we need to instrument;
      instrument_coll_const(breakpointsymb, file);
      
    } else {
      // for tracing and profiling;
      instrument_coll_const(profsymb, file);
    }


    if ( verboseMode ) printf ("\tgenerating dep file %s\n", outFile);
    file->saveDepFile(outFile);
  }
}


void instrument_coll_const (SgSymbol * fsymb, SgFile *file) {

    /* fsymb is function symbol to insert at collection construction points */

    /* -- Instrument all collection variable declarations          -- */
    /* -------------------------------------------------------------- */
    SgSymbol *classsymb;
    char *classname;
    SgFunctionCallExp *call;
    SgExprListExp *exp;
    SgCExpStmt *c;
    SgValueExp *v;
    SgType *elemType;
    SgDerivedCollectionType *d;
    char *elemname;
    SgVarDeclStmt *var;
    SgCExpStmt *ass;
    SgSymbol *varsymb;
    SgExprListExp *elist;
    SgStatement *s;

    s = file->firstStatement();
    while ( s )
    {
      if ( var = isSgVarDeclStmt (s) )
      {
        for (int i=0; i<var->numberOfSymbols(); i++)
        {
          varsymb = var->symbol(i);
          if ( d = isSgDerivedCollectionType (varsymb->type()) )
          {
            /* -- Symbol is of type CollectionType<ElementType>    -- */
            /* -- get type names                                   -- */
            /* ------------------------------------------------------ */
            classsymb = d->collectionName();
            classname = classsymb->identifier();
            elemType  = d->elementClass();
            elemname  = (elemType->symbol())->identifier();
            if ( verboseMode )
              printf ("\tfound %s<%s> %s\n",
              classname, elemname, varsymb->identifier());

	    /* -- Declare and insert call to function passed in <fsymb> -- */
            /* -- pass all type and variable names                 -- */
            /* ------------------------------------------------------ */
            call = new SgFunctionCallExp(*fsymb);
            v = new SgValueExp(classname);
            call->addArg(*v);
            delete(v);
            v = new SgValueExp(elemname);
            call->addArg(*v);
            delete(v);
            v = new SgValueExp(varsymb->identifier());
            call->addArg(*v);
            delete(v);
  
            exp = new SgExprListExp();
            exp->setValue(*call);
            c = new SgCExpStmt (*exp);

            var->insertStmtBefore(*c);
          }
        }
      }
      else if ( (ass=isSgCExpStmt (s)) && (elist=isSgExprListExp(ass->expr())) )
      {
        SgExpression *e;
        SgNewExp *n;
        SgPointerType *t;
        char *vname;

        for (int i=0; i<elist->length(); i++)
        {
          e = elist->elem(i);
          if ( (e->variant() == ASSGN_OP) &&
               (n = isSgNewExp(e->rhs())) &&
               (t = isSgPointerType(n->type())) &&
               (d = isSgDerivedCollectionType(t->baseType()))
             )
          {
            /* -- dynamic collection allocation Statement:         -- */
            /* -- var = new CollectionType<ElementType>()          -- */
            /* ------------------------------------------------------ */
            classsymb = d->collectionName();
            classname = classsymb->identifier();
            elemType  = d->elementClass();
            elemname  = (elemType->symbol())->identifier();
            vname     = new char[strlen(e->lhs()->symbol()->identifier())+2];
            sprintf (vname, "*%s", e->lhs()->symbol()->identifier());

            if ( verboseMode )
              printf ("\tfound %s<%s> %s\n", classname, elemname, vname);

            /* -- Declare and insert call to PROFILE_COLL_NAMES    -- */
            /* -- pass all type and variable names                 -- */
            /* ------------------------------------------------------ */
            call = new SgFunctionCallExp(*fsymb);
            v = new SgValueExp(classname);
            call->addArg(*v);
            delete(v);
            v = new SgValueExp(elemname);
            call->addArg(*v);
            delete(v);
            v = new SgValueExp(vname);
            call->addArg(*v);
            delete(v);
  
            exp = new SgExprListExp();
            exp->setValue(*call);
            c = new SgCExpStmt (*exp);

            ass->insertStmtBefore(*c);
          }
        }
      }
      s = s->lexNext();
    }


  }
